@@ -1,7 +1,7 @@
 #define IN_LIBEXSLT
 #include "libexslt/libexslt.h"
 
-#if defined(WIN32) && !defined (__CYGWIN__) && (!__MINGW32__)
+#if defined(_WIN32) && !defined (__CYGWIN__) && (!__MINGW32__)
 #include <win32config.h>
 #else
 #include "config.h"
@@ -298,7 +298,7 @@ static void
 exsltMathLowestFunction (xmlXPathParserContextPtr ctxt, int nargs) {
     xmlNodeSetPtr ns, ret;
     void *user = NULL;
-    
+
 
     if (nargs != 1) {
 	xmlXPathSetArityError(ctxt);
@@ -368,7 +368,7 @@ exsltMathConstant (xmlChar *name, double precision) {
 
         if (precision <= len)
             len = (int)precision;
-        
+
         str = xmlStrsub(EXSLT_PI, 0, len);
 
     } else if (xmlStrEqual(name, BAD_CAST "E")) {
@@ -376,7 +376,7 @@ exsltMathConstant (xmlChar *name, double precision) {
 
         if (precision <= len)
             len = (int)precision;
-        
+
         str = xmlStrsub(EXSLT_E, 0, len);
 
     } else if (xmlStrEqual(name, BAD_CAST "SQRRT2")) {
@@ -384,7 +384,7 @@ exsltMathConstant (xmlChar *name, double precision) {
 
         if (precision <= len)
             len = (int)precision;
-        
+
         str = xmlStrsub(EXSLT_SQRRT2, 0, len);
 
     } else if (xmlStrEqual(name, BAD_CAST "LN2")) {
@@ -392,7 +392,7 @@ exsltMathConstant (xmlChar *name, double precision) {
 
         if (precision <= len)
             len = (int)precision;
-        
+
         str = xmlStrsub(EXSLT_LN2, 0, len);
 
     } else if (xmlStrEqual(name, BAD_CAST "LN10")) {
@@ -400,7 +400,7 @@ exsltMathConstant (xmlChar *name, double precision) {
 
         if (precision <= len)
             len = (int)precision;
-        
+
         str = xmlStrsub(EXSLT_LN10, 0, len);
 
     } else if (xmlStrEqual(name, BAD_CAST "LOG2E")) {
@@ -408,7 +408,7 @@ exsltMathConstant (xmlChar *name, double precision) {
 
         if (precision <= len)
             len = (int)precision;
-        
+
         str = xmlStrsub(EXSLT_LOG2E, 0, len);
 
     } else if (xmlStrEqual(name, BAD_CAST "SQRT1_2")) {
@@ -416,7 +416,7 @@ exsltMathConstant (xmlChar *name, double precision) {
 
         if (precision <= len)
             len = (int)precision;
-        
+
         str = xmlStrsub(EXSLT_SQRT1_2, 0, len);
 
     } else {
@@ -972,22 +972,22 @@ exsltMathAtan2 (double y, double x) {
  */
 static void
 exsltMathAtan2Function (xmlXPathParserContextPtr ctxt, int nargs) {
-    double ret, y;
+    double ret, x;
 
     if (nargs != 2) {
 	xmlXPathSetArityError(ctxt);
 	return;
     }
-    y = xmlXPathPopNumber(ctxt);
+    x = xmlXPathPopNumber(ctxt);
     if (xmlXPathCheckError(ctxt))
 	return;
 
-    /* x */
+    /* y */
     ret = xmlXPathPopNumber(ctxt);
     if (xmlXPathCheckError(ctxt))
 	return;
 
-    ret = exsltMathAtan2(y, ret);
+    ret = exsltMathAtan2(ret, x);
 
     xmlXPathReturnNumber(ctxt, ret);
 }
@@ -1106,3 +1106,97 @@ exsltMathRegister (void) {
 				   exsltMathExpFunction);
 #endif
 }
+
+/**
+ * exsltMathXpathCtxtRegister:
+ *
+ * Registers the EXSLT - Math module for use outside XSLT
+ */
+int
+exsltMathXpathCtxtRegister (xmlXPathContextPtr ctxt, const xmlChar *prefix)
+{
+    if (ctxt
+        && prefix
+        && !xmlXPathRegisterNs(ctxt,
+                               prefix,
+                               (const xmlChar *) EXSLT_MATH_NAMESPACE)
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "min",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathMinFunction)
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "max",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathMaxFunction)
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "highest",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathHighestFunction)
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "lowest",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathLowestFunction)
+#ifdef HAVE_STDLIB_H
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "random",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathRandomFunction)
+#endif
+#if HAVE_MATH_H
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "abs",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathAbsFunction)
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "sqrt",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathSqrtFunction)
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "power",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathPowerFunction)
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "log",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathLogFunction)
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "sin",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathSinFunction)
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "cos",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathCosFunction)
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "tan",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathTanFunction)
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "asin",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathAsinFunction)
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "acos",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathAcosFunction)
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "atan",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathAtanFunction)
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "atan2",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathAtan2Function)
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "exp",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathExpFunction)
+#endif
+        && !xmlXPathRegisterFuncNS(ctxt,
+                                   (const xmlChar *) "constant",
+                                   (const xmlChar *) EXSLT_MATH_NAMESPACE,
+                                   exsltMathConstantFunction)) {
+        return 0;
+    }
+    return -1;
+}
