@@ -34,17 +34,19 @@
 
 #ifdef _WIN32
 #include <stdlib.h>             /* for _MAX_PATH */
+#ifndef PATH_MAX
 #define PATH_MAX _MAX_PATH
 #endif
+#endif
 
 #ifdef WITH_XSLT_DEBUG
 #define WITH_XSLT_DEBUG_EXTENSIONS
 #endif
 
 /************************************************************************
- * 									*
- * 			Private Types and Globals			*
- * 									*
+ *									*
+ *			Private Types and Globals			*
+ *									*
  ************************************************************************/
 
 typedef struct _xsltExtDef xsltExtDef;
@@ -84,11 +86,12 @@ static xmlHashTablePtr xsltFunctionsHash = NULL;
 static xmlHashTablePtr xsltElementsHash = NULL;
 static xmlHashTablePtr xsltTopLevelsHash = NULL;
 static xmlHashTablePtr xsltModuleHash = NULL;
+static xmlMutexPtr xsltExtMutex = NULL;
 
 /************************************************************************
- * 									*
- * 			Type functions 					*
- * 									*
+ *									*
+ *			Type functions					*
+ *									*
  ************************************************************************/
 
 /**
@@ -201,6 +204,11 @@ xsltFreeExtModule(xsltExtModulePtr ext)
     xmlFree(ext);
 }
 
+static void
+xsltFreeExtModuleEntry(void *payload, const xmlChar *name ATTRIBUTE_UNUSED) {
+    xsltFreeExtModule((xsltExtModulePtr) payload);
+}
+
 /**
  * xsltNewExtData:
  * @extModule:  the module
@@ -242,6 +250,11 @@ xsltFreeExtData(xsltExtDataPtr ext)
     xmlFree(ext);
 }
 
+static void
+xsltFreeExtDataEntry(void *payload, const xmlChar *name ATTRIBUTE_UNUSED) {
+    xsltFreeExtData((xsltExtDataPtr) payload);
+}
+
 /**
  * xsltNewExtElement:
  * @precomp:  the pre-computation function
@@ -286,6 +299,11 @@ xsltFreeExtElement(xsltExtElementPtr ext)
     xmlFree(ext);
 }
 
+static void
+xsltFreeExtElementEntry(void *payload, const xmlChar *name ATTRIBUTE_UNUSED) {
+    xsltFreeExtElement((xsltExtElementPtr) payload);
+}
+
 
 #ifdef WITH_MODULES
 typedef void (*exsltRegisterFunction) (void);
@@ -299,18 +317,18 @@ typedef void (*exsltRegisterFunction) (void);
  * @URI:  the function or element namespace URI
  *
  * Dynamically loads an extension plugin when available.
- * 
- * The plugin name is derived from the URI by removing the 
+ *
+ * The plugin name is derived from the URI by removing the
  * initial protocol designation, e.g. "http://", then converting
  * the characters ".", "-", "/", and "\" into "_", the removing
  * any trailing "/", then concatenating LIBXML_MODULE_EXTENSION.
- * 
- * Plugins are loaded from the directory specified by the 
- * environment variable LIBXSLT_PLUGINS_PATH, or if NULL, 
+ *
+ * Plugins are loaded from the directory specified by the
+ * environment variable LIBXSLT_PLUGINS_PATH, or if NULL,
  * by LIBXSLT_DEFAULT_PLUGINS_PATH() which is determined at
  * compile time.
  *
- * Returns 0 if successful, -1 in case of error. 
+ * Returns 0 if successful, -1 in case of error.
  */
 
 static int
@@ -337,10 +355,14 @@ xsltExtModuleRegisterDynamic(const xmlChar * URI)
             return (-1);
     }
 
+    xmlMutexLock(xsltExtMutex);
+
     /* have we attempted to register this module already? */
     if (xmlHashLookup(xsltModuleHash, URI) != NULL) {
+        xmlMutexUnlock(xsltExtMutex);
         return (-1);
     }
+    xmlMutexUnlock(xsltExtMutex);
 
     /* transform extension namespace into a module name */
     protocol = xmlStrstr(URI, BAD_CAST "://");
@@ -360,30 +382,33 @@ xsltExtModuleRegisterDynamic(const xmlChar * URI)
         i++;
     }
 
-    if (*(i - 1) == '_')
+    /* Strip underscores from end of string. */
+    while (i > ext_name && *(i - 1) == '_') {
+        i--;
         *i = '\0';
+    }
 
     /* determine module directory */
     ext_directory = (xmlChar *) getenv("LIBXSLT_PLUGINS_PATH");
 
-#ifdef WITH_XSLT_DEBUG_EXTENSIONS
-    xsltGenericDebug(xsltGenericDebugContext,
-                     "LIBXSLT_PLUGINS_PATH is %s\n", ext_directory);
-#endif
-
-    if (NULL == ext_directory)
+    if (NULL == ext_directory) {
         ext_directory = BAD_CAST LIBXSLT_DEFAULT_PLUGINS_PATH();
-    if (NULL == ext_directory)
-        return (-1);
+	if (NULL == ext_directory)
+	  return (-1);
+    }
+#ifdef WITH_XSLT_DEBUG_EXTENSIONS
+    else
+      xsltGenericDebug(xsltGenericDebugContext,
+		       "LIBXSLT_PLUGINS_PATH is %s\n", ext_directory);
+#endif
 
     /* build the module filename, and confirm the module exists */
     xmlStrPrintf((xmlChar *) module_filename, sizeof(module_filename),
-                 BAD_CAST "%s/%s%s",
-                 ext_directory, ext_name, LIBXML_MODULE_EXTENSION);
+                 "%s/%s%s", ext_directory, ext_name, LIBXML_MODULE_EXTENSION);
 
 #ifdef WITH_XSLT_DEBUG_EXTENSIONS
     xsltGenericDebug(xsltGenericDebugContext,
-                     "Attempting to load plugin: %s for URI: %s\n", 
+                     "Attempting to load plugin: %s for URI: %s\n",
                      module_filename, URI);
 #endif
 
@@ -427,12 +452,14 @@ xsltExtModuleRegisterDynamic(const xmlChar * URI)
         (*regfunc) ();
 
         /* register this module in our hash */
+        xmlMutexLock(xsltExtMutex);
         xmlHashAddEntry(xsltModuleHash, URI, (void *) m);
+        xmlMutexUnlock(xsltExtMutex);
     } else {
 
 #ifdef WITH_XSLT_DEBUG_EXTENSIONS
 	xsltGenericDebug(xsltGenericDebugContext,
-                     "xmlModuleSymbol failed for plugin: %s, regfunc: %s\n", 
+                     "xmlModuleSymbol failed for plugin: %s, regfunc: %s\n",
                      module_filename, regfunc_name);
 #endif
 
@@ -446,16 +473,16 @@ xsltExtModuleRegisterDynamic(const xmlChar * URI)
 }
 #else
 static int
-xsltExtModuleRegisterDynamic(const xmlChar * ATTRIBUTE_UNUSED URI)
+xsltExtModuleRegisterDynamic(const xmlChar * URI ATTRIBUTE_UNUSED)
 {
   return -1;
 }
 #endif
 
 /************************************************************************
- * 									*
- * 		The stylesheet extension prefixes handling		*
- * 									*
+ *									*
+ *		The stylesheet extension prefixes handling		*
+ *									*
  ************************************************************************/
 
 
@@ -477,7 +504,7 @@ xsltFreeExts(xsltStylesheetPtr style)
  * @style: an XSLT stylesheet
  * @prefix: the prefix used (optional)
  * @URI: the URI associated to the extension
- * 
+ *
  * Registers an extension namespace
  * This is called from xslt.c during compile-time.
  * The given prefix is not needed.
@@ -538,10 +565,14 @@ xsltRegisterExtPrefix(xsltStylesheetPtr style,
     if (xsltExtensionsHash != NULL) {
         xsltExtModulePtr module;
 
+        xmlMutexLock(xsltExtMutex);
         module = xmlHashLookup(xsltExtensionsHash, URI);
+        xmlMutexUnlock(xsltExtMutex);
         if (NULL == module) {
             if (!xsltExtModuleRegisterDynamic(URI)) {
+                xmlMutexLock(xsltExtMutex);
                 module = xmlHashLookup(xsltExtensionsHash, URI);
+                xmlMutexUnlock(xsltExtMutex);
             }
         }
         if (module != NULL) {
@@ -553,9 +584,9 @@ xsltRegisterExtPrefix(xsltStylesheetPtr style,
 }
 
 /************************************************************************
- * 									*
- * 		The extensions modules interfaces			*
- * 									*
+ *									*
+ *		The extensions modules interfaces			*
+ *									*
  ************************************************************************/
 
 /**
@@ -563,7 +594,7 @@ xsltRegisterExtPrefix(xsltStylesheetPtr style,
  * @ctxt: an XSLT transformation context
  * @name: the name of the element
  * @URI: the URI associated to the element
- * @function: the actual implementation which should be called 
+ * @function: the actual implementation which should be called
  *
  * Registers an extension function
  *
@@ -573,6 +604,8 @@ int
 xsltRegisterExtFunction(xsltTransformContextPtr ctxt, const xmlChar * name,
                         const xmlChar * URI, xmlXPathFunction function)
 {
+    int ret;
+
     if ((ctxt == NULL) || (name == NULL) ||
         (URI == NULL) || (function == NULL))
         return (-1);
@@ -583,8 +616,11 @@ xsltRegisterExtFunction(xsltTransformContextPtr ctxt, const xmlChar * name,
         ctxt->extFunctions = xmlHashCreate(10);
     if (ctxt->extFunctions == NULL)
         return (-1);
-    return (xmlHashAddEntry2
-            (ctxt->extFunctions, name, URI, XML_CAST_FPTR(function)));
+
+    ret = xmlHashAddEntry2(ctxt->extFunctions, name, URI,
+                           XML_CAST_FPTR(function));
+
+    return(ret);
 }
 
 /**
@@ -592,7 +628,7 @@ xsltRegisterExtFunction(xsltTransformContextPtr ctxt, const xmlChar * name,
  * @ctxt: an XSLT transformation context
  * @name: the name of the element
  * @URI: the URI associated to the element
- * @function: the actual implementation which should be called 
+ * @function: the actual implementation which should be called
  *
  * Registers an extension element
  *
@@ -647,8 +683,8 @@ xsltStyleInitializeStylesheetModule(xsltStylesheetPtr style,
     xsltExtDataPtr dataContainer;
     void *userData = NULL;
     xsltExtModulePtr module;
-    
-    if ((style == NULL) || (URI == NULL))	
+
+    if ((style == NULL) || (URI == NULL))
 	return(NULL);
 
     if (xsltExtensionsHash == NULL) {
@@ -659,7 +695,12 @@ xsltStyleInitializeStylesheetModule(xsltStylesheetPtr style,
 	return(NULL);
     }
 
+    xmlMutexLock(xsltExtMutex);
+
     module = xmlHashLookup(xsltExtensionsHash, URI);
+
+    xmlMutexUnlock(xsltExtMutex);
+
     if (module == NULL) {
 #ifdef WITH_XSLT_DEBUG_EXTENSIONS
 	xsltGenericDebug(xsltGenericDebugContext,
@@ -692,7 +733,7 @@ xsltStyleInitializeStylesheetModule(xsltStylesheetPtr style,
 	* Fire the initialization callback.
 	*/
 	userData = module->styleInitFunc(style, URI);
-    }    
+    }
     /*
     * Store the user-data in the context of the given stylesheet.
     */
@@ -703,7 +744,7 @@ xsltStyleInitializeStylesheetModule(xsltStylesheetPtr style,
     if (xmlHashAddEntry(style->extInfos, URI,
 	(void *) dataContainer) < 0)
     {
-	xsltTransformError(NULL, style, NULL,	    
+	xsltTransformError(NULL, style, NULL,
 	    "Failed to register module '%s'.\n", URI);
 	style->errors++;
 	if (module->styleShutdownFunc)
@@ -738,7 +779,7 @@ xsltStyleGetExtData(xsltStylesheetPtr style, const xmlChar * URI)
 	(xsltExtensionsHash == NULL))
 	return (NULL);
 
-    
+
 #ifdef XSLT_REFACTORED
     /*
     * This is intended for global storage, so only the main
@@ -814,7 +855,7 @@ xsltStyleStylesheetLevelGetExtData(xsltStylesheetPtr style,
 	*/
 	if (dataContainer)
 	    return(dataContainer->extData);
-    }  
+    }
 
     dataContainer =
         xsltStyleInitializeStylesheetModule(style, URI);
@@ -853,7 +894,12 @@ xsltGetExtData(xsltTransformContextPtr ctxt, const xmlChar * URI)
         void *extData;
         xsltExtModulePtr module;
 
+        xmlMutexLock(xsltExtMutex);
+
         module = xmlHashLookup(xsltExtensionsHash, URI);
+
+        xmlMutexUnlock(xsltExtMutex);
+
         if (module == NULL) {
 #ifdef WITH_XSLT_DEBUG_EXTENSIONS
             xsltGenericDebug(xsltGenericDebugContext,
@@ -905,9 +951,10 @@ struct _xsltInitExtCtxt {
  * Initializes an extension module
  */
 static void
-xsltInitCtxtExt(xsltExtDataPtr styleData, xsltInitExtCtxt * ctxt,
-                const xmlChar * URI)
+xsltInitCtxtExt(void *payload, void *data, const xmlChar * URI)
 {
+    xsltExtDataPtr styleData = (xsltExtDataPtr) payload;
+    xsltInitExtCtxt *ctxt = (xsltInitExtCtxt *) data;
     xsltExtModulePtr module;
     xsltExtDataPtr ctxtData;
     void *extData;
@@ -1000,8 +1047,7 @@ xsltInitCtxtExts(xsltTransformContextPtr ctxt)
 
     while (style != NULL) {
         if (style->extInfos != NULL) {
-            xmlHashScan(style->extInfos,
-                        (xmlHashScanner) xsltInitCtxtExt, &ctx);
+            xmlHashScan(style->extInfos, xsltInitCtxtExt, &ctx);
             if (ctx.ret == -1)
                 return (-1);
         }
@@ -1023,9 +1069,10 @@ xsltInitCtxtExts(xsltTransformContextPtr ctxt)
  * Shutdown an extension module loaded
  */
 static void
-xsltShutdownCtxtExt(xsltExtDataPtr data, xsltTransformContextPtr ctxt,
-                    const xmlChar * URI)
+xsltShutdownCtxtExt(void *payload, void *vctxt, const xmlChar * URI)
 {
+    xsltExtDataPtr data = (xsltExtDataPtr) payload;
+    xsltTransformContextPtr ctxt = (xsltTransformContextPtr) vctxt;
     xsltExtModulePtr module;
 
     if ((data == NULL) || (ctxt == NULL) || (URI == NULL))
@@ -1054,9 +1101,8 @@ xsltShutdownCtxtExts(xsltTransformContextPtr ctxt)
         return;
     if (ctxt->extInfos == NULL)
         return;
-    xmlHashScan(ctxt->extInfos, (xmlHashScanner) xsltShutdownCtxtExt,
-                ctxt);
-    xmlHashFree(ctxt->extInfos, (xmlHashDeallocator) xsltFreeExtData);
+    xmlHashScan(ctxt->extInfos, xsltShutdownCtxtExt, ctxt);
+    xmlHashFree(ctxt->extInfos, xsltFreeExtDataEntry);
     ctxt->extInfos = NULL;
 }
 
@@ -1069,9 +1115,10 @@ xsltShutdownCtxtExts(xsltTransformContextPtr ctxt)
  * Shutdown an extension module loaded
  */
 static void
-xsltShutdownExt(xsltExtDataPtr data, xsltStylesheetPtr style,
-                const xmlChar * URI)
+xsltShutdownExt(void *payload, void *vctxt, const xmlChar * URI)
 {
+    xsltExtDataPtr data = (xsltExtDataPtr) payload;
+    xsltStylesheetPtr style = (xsltStylesheetPtr) vctxt;
     xsltExtModulePtr module;
 
     if ((data == NULL) || (style == NULL) || (URI == NULL))
@@ -1089,9 +1136,8 @@ xsltShutdownExt(xsltExtDataPtr data, xsltStylesheetPtr style,
     * Don't remove the entry from the hash table here, since
     * this will produce segfaults - this fixes bug #340624.
     *
-    * xmlHashRemoveEntry(style->extInfos, URI,
-    *   (xmlHashDeallocator) xsltFreeExtData);
-    */    
+    * xmlHashRemoveEntry(style->extInfos, URI, xsltFreeExtDataEntry);
+    */
 }
 
 /**
@@ -1107,15 +1153,15 @@ xsltShutdownExts(xsltStylesheetPtr style)
         return;
     if (style->extInfos == NULL)
         return;
-    xmlHashScan(style->extInfos, (xmlHashScanner) xsltShutdownExt, style);
-    xmlHashFree(style->extInfos, (xmlHashDeallocator) xsltFreeExtData);
+    xmlHashScan(style->extInfos, xsltShutdownExt, style);
+    xmlHashFree(style->extInfos, xsltFreeExtDataEntry);
     style->extInfos = NULL;
 }
 
 /**
  * xsltCheckExtPrefix:
  * @style: the stylesheet
- * @URI: the namespace URI (possibly NULL)
+ * @URI: the namespace prefix (possibly NULL)
  *
  * Check if the given prefix is one of the declared extensions.
  * This is intended to be called only at compile-time.
@@ -1127,12 +1173,12 @@ xsltShutdownExts(xsltStylesheetPtr style)
  */
 int
 xsltCheckExtPrefix(xsltStylesheetPtr style, const xmlChar * URI)
-{    
+{
 #ifdef XSLT_REFACTORED
     if ((style == NULL) || (style->compCtxt == NULL) ||
 	(XSLT_CCTXT(style)->inode == NULL) ||
 	(XSLT_CCTXT(style)->inode->extElemNs == NULL))
-        return (0);    
+        return (0);
     /*
     * Lookup the extension namespaces registered
     * at the current node in the stylesheet's tree.
@@ -1146,7 +1192,7 @@ xsltCheckExtPrefix(xsltStylesheetPtr style, const xmlChar * URI)
 		URI))
 	    {
 		return(1);
-	    }	    
+	    }
 	}
     }
 #else
@@ -1172,6 +1218,37 @@ xsltCheckExtPrefix(xsltStylesheetPtr style, const xmlChar * URI)
     return (0);
 }
 
+/**
+ * xsltCheckExtURI:
+ * @style: the stylesheet
+ * @URI: the namespace URI (possibly NULL)
+ *
+ * Check if the given prefix is one of the declared extensions.
+ * This is intended to be called only at compile-time.
+ * Called by:
+ *  xsltPrecomputeStylesheet() (xslt.c)
+ *  xsltParseTemplateContent (xslt.c)
+ *
+ * Returns 1 if this is an extension, 0 otherwise
+ */
+int
+xsltCheckExtURI(xsltStylesheetPtr style, const xmlChar * URI)
+{
+    xsltExtDefPtr cur;
+
+    if ((style == NULL) || (style->nsDefs == NULL))
+        return (0);
+    if (URI == NULL)
+        return (0);
+    cur = (xsltExtDefPtr) style->nsDefs;
+    while (cur != NULL) {
+        if (xmlStrEqual(URI, cur->URI))
+            return (1);
+        cur = cur->next;
+    }
+    return (0);
+}
+
 /**
  * xsltRegisterExtModuleFull:
  * @URI:  URI associated to this module
@@ -1202,18 +1279,27 @@ xsltRegisterExtModuleFull(const xmlChar * URI,
     if (xsltExtensionsHash == NULL)
         return (-1);
 
+    xmlMutexLock(xsltExtMutex);
+
     module = xmlHashLookup(xsltExtensionsHash, URI);
     if (module != NULL) {
         if ((module->initFunc == initFunc) &&
             (module->shutdownFunc == shutdownFunc))
-            return (0);
-        return (-1);
+            ret = 0;
+        else
+            ret = -1;
+        goto done;
     }
     module = xsltNewExtModule(initFunc, shutdownFunc,
                               styleInitFunc, styleShutdownFunc);
-    if (module == NULL)
-        return (-1);
+    if (module == NULL) {
+        ret = -1;
+        goto done;
+    }
     ret = xmlHashAddEntry(xsltExtensionsHash, URI, (void *) module);
+
+done:
+    xmlMutexUnlock(xsltExtMutex);
     return (ret);
 }
 
@@ -1254,9 +1340,12 @@ xsltUnregisterExtModule(const xmlChar * URI)
     if (xsltExtensionsHash == NULL)
         return (-1);
 
-    ret =
-        xmlHashRemoveEntry(xsltExtensionsHash, URI,
-                           (xmlHashDeallocator) xsltFreeExtModule);
+    xmlMutexLock(xsltExtMutex);
+
+    ret = xmlHashRemoveEntry(xsltExtensionsHash, URI, xsltFreeExtModuleEntry);
+
+    xmlMutexUnlock(xsltExtMutex);
+
     return (ret);
 }
 
@@ -1271,9 +1360,12 @@ xsltUnregisterAllExtModules(void)
     if (xsltExtensionsHash == NULL)
         return;
 
-    xmlHashFree(xsltExtensionsHash,
-                (xmlHashDeallocator) xsltFreeExtModule);
+    xmlMutexLock(xsltExtMutex);
+
+    xmlHashFree(xsltExtensionsHash, xsltFreeExtModuleEntry);
     xsltExtensionsHash = NULL;
+
+    xmlMutexUnlock(xsltExtMutex);
 }
 
 /**
@@ -1318,9 +1410,13 @@ xsltRegisterExtModuleFunction(const xmlChar * name, const xmlChar * URI,
     if (xsltFunctionsHash == NULL)
         return (-1);
 
+    xmlMutexLock(xsltExtMutex);
+
     xmlHashUpdateEntry2(xsltFunctionsHash, name, URI,
                         XML_CAST_FPTR(function), NULL);
 
+    xmlMutexUnlock(xsltExtMutex);
+
     return (0);
 }
 
@@ -1341,13 +1437,21 @@ xsltExtModuleFunctionLookup(const xmlChar * name, const xmlChar * URI)
     if ((xsltFunctionsHash == NULL) || (name == NULL) || (URI == NULL))
         return (NULL);
 
+    xmlMutexLock(xsltExtMutex);
+
     XML_CAST_FPTR(ret) = xmlHashLookup2(xsltFunctionsHash, name, URI);
 
+    xmlMutexUnlock(xsltExtMutex);
+
     /* if lookup fails, attempt a dynamic load on supported platforms */
     if (NULL == ret) {
         if (!xsltExtModuleRegisterDynamic(URI)) {
+            xmlMutexLock(xsltExtMutex);
+
             XML_CAST_FPTR(ret) =
                 xmlHashLookup2(xsltFunctionsHash, name, URI);
+
+            xmlMutexUnlock(xsltExtMutex);
         }
     }
 
@@ -1366,10 +1470,18 @@ xsltExtModuleFunctionLookup(const xmlChar * name, const xmlChar * URI)
 int
 xsltUnregisterExtModuleFunction(const xmlChar * name, const xmlChar * URI)
 {
+    int ret;
+
     if ((xsltFunctionsHash == NULL) || (name == NULL) || (URI == NULL))
         return (-1);
 
-    return xmlHashRemoveEntry2(xsltFunctionsHash, name, URI, NULL);
+    xmlMutexLock(xsltExtMutex);
+
+    ret = xmlHashRemoveEntry2(xsltFunctionsHash, name, URI, NULL);
+
+    xmlMutexUnlock(xsltExtMutex);
+
+    return(ret);
 }
 
 /**
@@ -1380,11 +1492,20 @@ xsltUnregisterExtModuleFunction(const xmlChar * name, const xmlChar * URI)
 static void
 xsltUnregisterAllExtModuleFunction(void)
 {
+    xmlMutexLock(xsltExtMutex);
+
     xmlHashFree(xsltFunctionsHash, NULL);
     xsltFunctionsHash = NULL;
+
+    xmlMutexUnlock(xsltExtMutex);
 }
 
 
+static void
+xsltFreeElemPreComp(xsltElemPreCompPtr comp) {
+    xmlFree(comp);
+}
+
 /**
  * xsltNewElemPreComp:
  * @style:  the XSLT stylesheet
@@ -1409,8 +1530,7 @@ xsltNewElemPreComp(xsltStylesheetPtr style, xmlNodePtr inst,
     }
     memset(cur, 0, sizeof(xsltElemPreComp));
 
-    xsltInitElemPreComp(cur, style, inst, function,
-                        (xsltElemPreCompDeallocator) xmlFree);
+    xsltInitElemPreComp(cur, style, inst, function, xsltFreeElemPreComp);
 
     return (cur);
 }
@@ -1461,8 +1581,13 @@ xsltPreComputeExtModuleElement(xsltStylesheetPtr style, xmlNodePtr inst)
         (inst->type != XML_ELEMENT_NODE) || (inst->ns == NULL))
         return (NULL);
 
+    xmlMutexLock(xsltExtMutex);
+
     ext = (xsltExtElementPtr)
         xmlHashLookup2(xsltElementsHash, inst->name, inst->ns->href);
+
+    xmlMutexUnlock(xsltExtMutex);
+
     /*
     * EXT TODO: Now what?
     */
@@ -1516,6 +1641,8 @@ xsltRegisterExtModuleElement(const xmlChar * name, const xmlChar * URI,
                              xsltPreComputeFunction precomp,
                              xsltTransformFunction transform)
 {
+    int ret = 0;
+
     xsltExtElementPtr ext;
 
     if ((name == NULL) || (URI == NULL) || (transform == NULL))
@@ -1526,14 +1653,21 @@ xsltRegisterExtModuleElement(const xmlChar * name, const xmlChar * URI,
     if (xsltElementsHash == NULL)
         return (-1);
 
+    xmlMutexLock(xsltExtMutex);
+
     ext = xsltNewExtElement(precomp, transform);
-    if (ext == NULL)
-        return (-1);
+    if (ext == NULL) {
+        ret = -1;
+        goto done;
+    }
 
     xmlHashUpdateEntry2(xsltElementsHash, name, URI, (void *) ext,
-                        (xmlHashDeallocator) xsltFreeExtElement);
+                        xsltFreeExtElementEntry);
 
-    return (0);
+done:
+    xmlMutexUnlock(xsltExtMutex);
+
+    return (ret);
 }
 
 /**
@@ -1558,10 +1692,14 @@ xsltExtElementLookup(xsltTransformContextPtr ctxt,
 
     if ((ctxt != NULL) && (ctxt->extElements != NULL)) {
         XML_CAST_FPTR(ret) = xmlHashLookup2(ctxt->extElements, name, URI);
-        if (ret != NULL)
-            return (ret);
+        if (ret != NULL) {
+            return(ret);
+        }
     }
-    return xsltExtModuleElementLookup(name, URI);
+
+    ret = xsltExtModuleElementLookup(name, URI);
+
+    return (ret);
 }
 
 /**
@@ -1581,14 +1719,24 @@ xsltExtModuleElementLookup(const xmlChar * name, const xmlChar * URI)
     if ((xsltElementsHash == NULL) || (name == NULL) || (URI == NULL))
         return (NULL);
 
+    xmlMutexLock(xsltExtMutex);
+
     ext = (xsltExtElementPtr) xmlHashLookup2(xsltElementsHash, name, URI);
 
-    /* if function lookup fails, attempt a dynamic load on supported platforms */
-    ext = (xsltExtElementPtr) xmlHashLookup2(xsltElementsHash, name, URI);
+    xmlMutexUnlock(xsltExtMutex);
+
+    /*
+     * if function lookup fails, attempt a dynamic load on
+     * supported platforms
+     */
     if (NULL == ext) {
         if (!xsltExtModuleRegisterDynamic(URI)) {
+            xmlMutexLock(xsltExtMutex);
+
             ext = (xsltExtElementPtr)
 	          xmlHashLookup2(xsltElementsHash, name, URI);
+
+            xmlMutexUnlock(xsltExtMutex);
         }
     }
 
@@ -1615,12 +1763,20 @@ xsltExtModuleElementPreComputeLookup(const xmlChar * name,
     if ((xsltElementsHash == NULL) || (name == NULL) || (URI == NULL))
         return (NULL);
 
+    xmlMutexLock(xsltExtMutex);
+
     ext = (xsltExtElementPtr) xmlHashLookup2(xsltElementsHash, name, URI);
 
+    xmlMutexUnlock(xsltExtMutex);
+
     if (ext == NULL) {
         if (!xsltExtModuleRegisterDynamic(URI)) {
+            xmlMutexLock(xsltExtMutex);
+
             ext = (xsltExtElementPtr)
 	          xmlHashLookup2(xsltElementsHash, name, URI);
+
+            xmlMutexUnlock(xsltExtMutex);
         }
     }
 
@@ -1641,11 +1797,19 @@ xsltExtModuleElementPreComputeLookup(const xmlChar * name,
 int
 xsltUnregisterExtModuleElement(const xmlChar * name, const xmlChar * URI)
 {
+    int ret;
+
     if ((xsltElementsHash == NULL) || (name == NULL) || (URI == NULL))
         return (-1);
 
-    return xmlHashRemoveEntry2(xsltElementsHash, name, URI,
-                               (xmlHashDeallocator) xsltFreeExtElement);
+    xmlMutexLock(xsltExtMutex);
+
+    ret = xmlHashRemoveEntry2(xsltElementsHash, name, URI,
+                              xsltFreeExtElementEntry);
+
+    xmlMutexUnlock(xsltExtMutex);
+
+    return(ret);
 }
 
 /**
@@ -1656,8 +1820,12 @@ xsltUnregisterExtModuleElement(const xmlChar * name, const xmlChar * URI)
 static void
 xsltUnregisterAllExtModuleElement(void)
 {
-    xmlHashFree(xsltElementsHash, (xmlHashDeallocator) xsltFreeExtElement);
+    xmlMutexLock(xsltExtMutex);
+
+    xmlHashFree(xsltElementsHash, xsltFreeExtElementEntry);
     xsltElementsHash = NULL;
+
+    xmlMutexUnlock(xsltExtMutex);
 }
 
 /**
@@ -1682,9 +1850,13 @@ xsltRegisterExtModuleTopLevel(const xmlChar * name, const xmlChar * URI,
     if (xsltTopLevelsHash == NULL)
         return (-1);
 
+    xmlMutexLock(xsltExtMutex);
+
     xmlHashUpdateEntry2(xsltTopLevelsHash, name, URI,
                         XML_CAST_FPTR(function), NULL);
 
+    xmlMutexUnlock(xsltExtMutex);
+
     return (0);
 }
 
@@ -1705,12 +1877,20 @@ xsltExtModuleTopLevelLookup(const xmlChar * name, const xmlChar * URI)
     if ((xsltTopLevelsHash == NULL) || (name == NULL) || (URI == NULL))
         return (NULL);
 
+    xmlMutexLock(xsltExtMutex);
+
     XML_CAST_FPTR(ret) = xmlHashLookup2(xsltTopLevelsHash, name, URI);
 
+    xmlMutexUnlock(xsltExtMutex);
+
     /* if lookup fails, attempt a dynamic load on supported platforms */
     if (NULL == ret) {
         if (!xsltExtModuleRegisterDynamic(URI)) {
+            xmlMutexLock(xsltExtMutex);
+
             XML_CAST_FPTR(ret) = xmlHashLookup2(xsltTopLevelsHash, name, URI);
+
+            xmlMutexUnlock(xsltExtMutex);
         }
     }
 
@@ -1729,10 +1909,18 @@ xsltExtModuleTopLevelLookup(const xmlChar * name, const xmlChar * URI)
 int
 xsltUnregisterExtModuleTopLevel(const xmlChar * name, const xmlChar * URI)
 {
+    int ret;
+
     if ((xsltTopLevelsHash == NULL) || (name == NULL) || (URI == NULL))
         return (-1);
 
-    return xmlHashRemoveEntry2(xsltTopLevelsHash, name, URI, NULL);
+    xmlMutexLock(xsltExtMutex);
+
+    ret = xmlHashRemoveEntry2(xsltTopLevelsHash, name, URI, NULL);
+
+    xmlMutexUnlock(xsltExtMutex);
+
+    return(ret);
 }
 
 /**
@@ -1743,8 +1931,12 @@ xsltUnregisterExtModuleTopLevel(const xmlChar * name, const xmlChar * URI)
 static void
 xsltUnregisterAllExtModuleTopLevel(void)
 {
+    xmlMutexLock(xsltExtMutex);
+
     xmlHashFree(xsltTopLevelsHash, NULL);
     xsltTopLevelsHash = NULL;
+
+    xmlMutexUnlock(xsltExtMutex);
 }
 
 /**
@@ -1777,15 +1969,15 @@ xsltGetExtInfo(xsltStylesheetPtr style, const xmlChar * URI)
 }
 
 /************************************************************************
- * 									*
- * 		Test module http://xmlsoft.org/XSLT/			*
- * 									*
+ *									*
+ *		Test module http://xmlsoft.org/XSLT/			*
+ *									*
  ************************************************************************/
 
 /************************************************************************
- * 									*
- * 		Test of the extension module API			*
- * 									*
+ *									*
+ *		Test of the extension module API			*
+ *									*
  ************************************************************************/
 
 static xmlChar *testData = NULL;
@@ -1888,7 +2080,7 @@ xsltExtElementPreCompTest(xsltStylesheetPtr style, xmlNodePtr inst,
  * @ctxt:  an XSLT processing context
  * @node:  The current node
  * @inst:  the instruction in the stylesheet
- * @comp:  precomputed informations
+ * @comp:  precomputed information
  *
  * Process a libxslt:test node
  */
@@ -1951,7 +2143,7 @@ xsltExtInitTest(xsltTransformContextPtr ctxt, const xmlChar * URI)
         xsltGenericDebug(xsltGenericErrorContext,
                          "xsltExtInitTest: not initialized,"
                          " calling xsltStyleGetExtData\n");
-        xsltStyleGetExtData(ctxt->style, URI);
+        testStyleData = xsltStyleGetExtData(ctxt->style, URI);
         if (testStyleData == NULL) {
             xsltTransformError(ctxt, NULL, NULL,
                                "xsltExtInitTest: not initialized\n");
@@ -2055,6 +2247,7 @@ xsltExtStyleShutdownTest(xsltStylesheetPtr style ATTRIBUTE_UNUSED,
 void
 xsltRegisterTestModule(void)
 {
+    xsltInitGlobals();
     xsltRegisterExtModuleFull((const xmlChar *) XSLT_DEFAULT_URL,
                               xsltExtInitTest, xsltExtShutdownTest,
                               xsltExtStyleInitTest,
@@ -2069,14 +2262,28 @@ xsltRegisterTestModule(void)
 }
 
 static void
-xsltHashScannerModuleFree(void *payload, void *data ATTRIBUTE_UNUSED,
-                          xmlChar * name ATTRIBUTE_UNUSED)
+xsltHashScannerModuleFree(void *payload ATTRIBUTE_UNUSED,
+                          void *data ATTRIBUTE_UNUSED,
+                          const xmlChar *name ATTRIBUTE_UNUSED)
 {
 #ifdef WITH_MODULES
     xmlModuleClose(payload);
 #endif
 }
 
+/**
+ * xsltInitGlobals:
+ *
+ * Initialize the global variables for extensions
+ */
+void
+xsltInitGlobals(void)
+{
+    if (xsltExtMutex == NULL) {
+        xsltExtMutex = xmlNewMutex();
+    }
+}
+
 /**
  * xsltCleanupGlobals:
  *
@@ -2090,22 +2297,28 @@ xsltCleanupGlobals(void)
     xsltUnregisterAllExtModuleElement();
     xsltUnregisterAllExtModuleTopLevel();
 
+    xmlMutexLock(xsltExtMutex);
     /* cleanup dynamic module hash */
     if (NULL != xsltModuleHash) {
         xmlHashScan(xsltModuleHash, xsltHashScannerModuleFree, 0);
         xmlHashFree(xsltModuleHash, NULL);
         xsltModuleHash = NULL;
     }
+    xmlMutexUnlock(xsltExtMutex);
 
+    xmlFreeMutex(xsltExtMutex);
+    xsltExtMutex = NULL;
+    xsltFreeLocales();
     xsltUninit();
 }
 
 static void
 xsltDebugDumpExtensionsCallback(void *function ATTRIBUTE_UNUSED,
-                                FILE * output, const xmlChar * name,
+                                void *data, const xmlChar * name,
                                 const xmlChar * URI,
                                 const xmlChar * not_used ATTRIBUTE_UNUSED)
 {
+    FILE *output = (FILE *) data;
     if (!name || !URI)
         return;
     fprintf(output, "{%s}%s\n", URI, name);
@@ -2113,10 +2326,11 @@ xsltDebugDumpExtensionsCallback(void *function ATTRIBUTE_UNUSED,
 
 static void
 xsltDebugDumpExtModulesCallback(void *function ATTRIBUTE_UNUSED,
-                                FILE * output, const xmlChar * URI,
+                                void *data, const xmlChar * URI,
                                 const xmlChar * not_used ATTRIBUTE_UNUSED,
                                 const xmlChar * not_used2 ATTRIBUTE_UNUSED)
 {
+    FILE *output = (FILE *) data;
     if (!URI)
         return;
     fprintf(output, "%s\n", URI);
@@ -2139,25 +2353,28 @@ xsltDebugDumpExtensions(FILE * output)
         fprintf(output, "No registered extension functions\n");
     else {
         fprintf(output, "Registered Extension Functions:\n");
-        xmlHashScanFull(xsltFunctionsHash,
-                        (xmlHashScannerFull)
-                        xsltDebugDumpExtensionsCallback, output);
+        xmlMutexLock(xsltExtMutex);
+        xmlHashScanFull(xsltFunctionsHash, xsltDebugDumpExtensionsCallback,
+                        output);
+        xmlMutexUnlock(xsltExtMutex);
     }
     if (!xsltElementsHash)
         fprintf(output, "\nNo registered extension elements\n");
     else {
         fprintf(output, "\nRegistered Extension Elements:\n");
-        xmlHashScanFull(xsltElementsHash,
-                        (xmlHashScannerFull)
-                        xsltDebugDumpExtensionsCallback, output);
+        xmlMutexLock(xsltExtMutex);
+        xmlHashScanFull(xsltElementsHash, xsltDebugDumpExtensionsCallback,
+                        output);
+        xmlMutexUnlock(xsltExtMutex);
     }
     if (!xsltExtensionsHash)
         fprintf(output, "\nNo registered extension modules\n");
     else {
         fprintf(output, "\nRegistered Extension Modules:\n");
-        xmlHashScanFull(xsltExtensionsHash,
-                        (xmlHashScannerFull)
-                        xsltDebugDumpExtModulesCallback, output);
+        xmlMutexLock(xsltExtMutex);
+        xmlHashScanFull(xsltExtensionsHash, xsltDebugDumpExtModulesCallback,
+                        output);
+        xmlMutexUnlock(xsltExtMutex);
     }
 
 }
